"""\
Purpose: Simulate a CSTA Application
Initial Version: Costas Skarakis 7/20/2020
"""
import os
import traceback
from _socket import timeout as sock_timeout
from threading import Lock, Thread, Event
from time import time, sleep

from common.client import TCPClient
from common.tc_logging import debug, warning, exception
from csta.CstaEndpoint import get_xml
from csta.CstaUser import CstaUser
from csta.CstaParser import parseBytes, buildMessageFromFile, buildMessage


def get_buffered_message(buffer, message, call_id=None, monitor_x_ref_id=None, calling_device=None):
    """
    Get a message from the specified buffer.
    Best used within a threading.Lock context

    :param buffer: the buffer to search
    :param message: the message_type to get
    :param call_id: the callID (if any) to request
    :param calling_device: used to separate csta events when needed
    :param monitor_x_ref_id: the monitorCrossRefID (if any) to request
    :return: the corresponding CstaMessage object
    """
    if message.endswith("Event"):
        # This will fail for first csta event
        key = str(monitor_x_ref_id) + str(message)
        if key in buffer:
            # message_index = None
            message_index = 0
            # for index in range(len(buffer[key])):
            #     # Csta Event could include multiple callID tags. Eg TransferredEvent
            #     call_id_list = buffer[key][index].get_all("callID")
            #     if str(call_id) in call_id_list or (not call_id and not call_id_list):
            #         message_index = index
            #         break
            if message_index is not None:
                buffered_message = buffer[key].pop(message_index)
                if not buffer[key]:
                    buffer.pop(key)
                return buffered_message

    # Here we look for first csta event, or csta responses
    for k in buffer:
        if ((message.endswith("Event") and call_id is None) or message.endswith("Response")) and \
                k.endswith(str(message)):
            buffered_message = buffer[k][0]

            if (calling_device is not None and
                    calling_device not in str(buffered_message).split("callingDevice")[1].split("callingDevice")[0]):
                continue
            else:
                buffer[k].pop(0)
                if not buffer[k]:
                    buffer.pop(k)
                return buffered_message


class CstaApplication:
    def __init__(self, server=False):
        self.ip = None
        self.port = None
        self.link = None
        self.min_event_id = 0
        self.users = {}
        self.parameters = {}
        self.server = server
        self.message_buffer = {}
        self.buffer_mod_time = None
        self.buffer_trigger_count = 0
        self.lock = Lock()
        self.buffer_event = Event()
        self.shutdown_flag = False
        self.waitForCstaMessage = self.wait_for_csta_message  # compatibility alias
        self.wait_thread = None

    def shutdown(self):
        """
        Try to stop threads and cleanup connections
        """
        self.shutdown_flag = True
        self.link.shutdown()
        self.wait_thread.join()

    def start_wait_thread(self):
        """
        Starts a separate thread that will consume incoming csta traffic and place it into buffers
        """
        self.wait_thread = Thread(target=self.wait_loop, daemon=True)
        self.wait_thread.start()

    def wait_loop(self):
        """
        Continuously wait for incoming csta messages. Add all messages to buffers
        """
        while not self.shutdown_flag:
            try:
                inbytes = self.link.waitForCstaData(timeout=None)
            except ConnectionError:
                # might need to add what kind of disconnection that was
                warning("Disconnected. Exiting")
                break
            if inbytes is None:
                warning("Disconnected. Will retry in 1 second")
                sleep(1)
            else:
                inmessage = parseBytes(inbytes)
                try:
                    net_object = self
                    for dn in self.users:
                        user = self.users[dn]
                        if (inmessage.is_event() and user.monitorCrossRefID == inmessage["monitorCrossRefID"]) or \
                                (inmessage.is_response() and inmessage.eventid in user.out_transactions):
                            net_object = user
                            break
                    with net_object.lock:
                        self.buffer_message(net_object, inmessage)
                except:
                    exception(traceback.format_exc())

    def connect(self, local_address, destination_address, protocol="tcp"):
        """ Connect to CSTA Server """
        local_ip, local_port = local_address
        self.ip = local_ip
        # params = {"source_ip": local_ip, "source_port": local_port, "transport": protocol}
        dest_ip, dest_port = destination_address
        # Only TCP implemented
        # 0 means bind to any available local port
        csta_link = TCPClient(local_ip, local_port)
        self.link = csta_link
        csta_link.connect(dest_ip, dest_port)
        self.port = csta_link.port

        inBytes = csta_link.waitForData()
        req = parseBytes(inBytes)

        resp = buildMessageFromFile(get_xml("SystemStatusResponse.xml"), {}, eventid=req.eventid)
        csta_link.send(resp.contents())

        reg = buildMessageFromFile(get_xml("SystemRegister.xml"), {}, eventid=0)
        csta_link.send(reg.contents())

        inBytes = csta_link.waitForData()
        reg_resp = parseBytes(inBytes)
        assert reg_resp.event == "SystemRegisterResponse", 'Invalid Response to System Register Request'
        self.start_wait_thread()
        return csta_link

    def new_user(self, directory_number):
        """ Add a new user to the Application
        :param directory_number: The users number as a string
        :return: The CstaUser object created for this user
        """
        directory_number = str(directory_number)
        if directory_number in self.users:
            warning(directory_number + " already exists")
            return self.users[directory_number]
        else:
            user = CstaUser(directory_number, self)
            self.users[directory_number] = user
            return user

    def get_user(self, directory_number):
        return self.users[str(directory_number)]

    def monitor_start(self, directory_number, force=False):
        """
        Send MonitorStart and add directory number to monitored users


        :param directory_number: The DN of the device to monitor
        :param force: Force sending another monitor request and create another Monitor Record for the same device
        :return: None
        """

        directory_number = str(directory_number)
        user = self.get_user(directory_number)
        if force or user.monitorCrossRefID is None:
            user.monitorCrossRefID = ""  # set this temporarily to avoid race condition sending monitor twice for a user
            user.parameters["user"] = directory_number
            try:
                user.send("MonitorStart")
                inmessage = user.wait_for_message("MonitorStartResponse")
                user.parameters["monitorCrossRefID"] = inmessage["monitorCrossRefID"]
                user.monitorCrossRefID = user.parameters["monitorCrossRefID"]
            except sock_timeout:
                user.monitorCrossRefID = None
        else:
            debug("Skip MonitorStart for already Monitored user: " + directory_number + ". Use force=True to override.")

    # def monitor_stop(self, directory_number):
    #     """ Send MonitorStop and delete directory number from monitored users"""
    #     csta_link = self.link
    #     m = buildMessageFromFile(get_xml("MonitorStop.xml"), {"user": directory_number}, eventid=1)
    #     csta_link.send(m.contents())
    #     in_bytes = csta_link.waitForData()
    #     inmessage = parseBytes(in_bytes)
    #     assert inmessage.event == "MonitorStopResponse", "Sent:{}  Received:{}".format(m.event, str(inmessage))
    #     self.parameters.pop(directory_number)

    def get_monitored_users(self):
        return list(user for user in self.users if self.users[str(user)].monitorCrossRefID is not None)

    def prepare_message(self, from_user, message):
        """
        Convert a SIP message string to SipMessage object

        :param message: The message string
        :param from_user: The originating CSTA subscriber object
        :return: A CstaMessage object
        """
        user = self.users[from_user]
        return buildMessage(message, user.parameters)

    def send(self, message, from_user=None, to_user=None, callID=None):
        """ Send a CSTA message

        :param callID: if provided, will use this callID in the outgoing message
        :param from_user: calling number
        :param to_user: called number
        :param message: This could be a predefined csta message or just the type of the message.
                       There are already several builtin xml files that are named after the corresponding messages
                       In order to access them just provide the type of the CSTA message.
                       There should be a matching filename in csta/xml

                       Example: C.send_csta("MakeCall")
        """
        if hasattr(to_user, "number"):
            to_user = to_user.number
        user = self.users[from_user]
        user.set_parameter("callingDevice", from_user)
        user.set_parameter("calledDirectoryNumber", to_user)
        if message.strip().startswith("<?xml"):
            m = buildMessage(message, user.parameters, eventid=0)
        else:
            m = buildMessageFromFile(get_xml(message), user.parameters, eventid=0)
        if from_user and m.event != "MonitorStart":
            assert user.monitorCrossRefID, "User {} must be monitored before sending messages".format(
                from_user)
        with self.lock:
            # must do this serially to avoid sending multiple requests with the same invoke id
            eventid = user.get_transaction_id(m.event)
            m.set_eventid(eventid)
            if m["deviceID"] and m["callID"]:
                m["deviceID"] = user.deviceID
                if callID:
                    m["callID"] = callID
                    user.callID = callID
                else:
                    m["callID"] = user.callID
            user.update_outgoing_transactions(m)
            user.update_call_id(m)
        # old_link = self.link
        # try:
        self.link.send(m.contents())
        # except IOError:
        #     while old_link is self.link:
        #         sleep(0.1)
        #     debug("Retransmitting {} after disconnection and reconnection".format(m.event))
        #     self.link.send(m.contents())

        return m

    def wait_for_csta_message(self, for_user, message, calling_device=None, ignore_messages=(), timeout=5.0):
        """
        Wait for CSTA message
        :param for_user: The message intended recipient
        :param message: The message to wait for
        :param calling_device: The calling device number. Used to separated csta events when necessary
        :param ignore_messages: Messages to ignore
        :param timeout: Defined timeout in seconds.
        :return: the CstaMessage received
        """
        callID = None
        # other_users_xrefid = [usr.parameters["monitorCrossRefID"]
        #                       if "monitorCrossRefID" in usr.parameters else None
        #                       for usr in self.users.values() if usr != for_user]
        # other_users_transactions = [usr.out_transactions for usr in self.users.values() if usr != for_user]
        user_xrefid = None
        this_user = None
        net_object = self
        if for_user is not None:
            this_user = self.users[for_user]
            # We are not using the below line, because if the same user is used, the xref id will be global
            # not thread specific
            # user_xrefid = this_user.parameters.get("monitorCrossRefID", None)
            user_xrefid = this_user.monitorCrossRefID
            net_object = this_user
            callID = this_user.parameters["callID"]
        # checked_buffer = None
        t0_tout = time()
        with net_object.lock:
            inmessage = get_buffered_message(net_object.message_buffer, message, callID, user_xrefid, calling_device)
        while not inmessage:
            # if not checked_buffer == net_object.buffer_mod_time:
            #     checked_buffer = net_object.buffer_mod_time
            if self.server:
                continue

            #            if not inmessage:
            rem_timeout = timeout - (time() - t0_tout)
            if rem_timeout > 0:
                net_object.buffer_event.wait(rem_timeout)
                net_object.buffer_event.clear()
                with net_object.lock:
                    inmessage = get_buffered_message(net_object.message_buffer, message, callID, user_xrefid,
                                                     calling_device)
                    if inmessage is None:
                        continue
            else:
                exception("%s (CSTA) No %s. Buffer lengths: %s. " % (for_user,
                                                                     message,
                                                                     str([(k, len(net_object.message_buffer[k]))
                                                                          for k in net_object.message_buffer])))
                raise sock_timeout

                # try:
                #     inbytes = self.link.waitForCstaData(timeout=timeout)
                #     inmessage = parseBytes(inbytes)

                # except UnicodeDecodeError:
                #     debug("Ignoring malformed data")
                #     inmessage = None
                #     continue
                # except sock_timeout:
                #     exception(str(for_user) + " No " + str(message) + " " + str(net_object.message_buffer))
                #     raise

            inmessage_type = inmessage.event
            if inmessage_type in ignore_messages:
                inmessage = None
                continue

            if message and (
                    # received message is not of expected type or
                    (isinstance(message, str) and message != inmessage_type) or

                    # received message type is not in list of expected types or
                    (type(message) in (list, tuple) and not any([m == inmessage_type for m in message]))
                    # or

                    # # received message is event and has an unexpected callID
                    # (inmessage.is_event() and
                    #  this_user is not None and
                    #  inmessage["callID"] and
                    #  callID and
                    #  callID != inmessage["callID"])  # or

                    # received message is a csta response that has an unknown/incorrect invokeID (eventid)
                    # (this_user is not None and
                    #  inmessage.is_response() and
                    #  inmessage.eventid not in this_user.out_transactions) or

                    # received message type is for another user (different xref_id and not MonitorStartResponse)
                    # # ie we are expecting message for a specific user
                    # (this_user is not None and
                    # # the message has a xrefid tag and our user has an xrefid parameter (although it may be empty)
                    # inmessage["monitorCrossRefID"] and user_xrefid is not None and
                    # # # the message xrefid is different than our user's xrefid
                    # inmessage["monitorCrossRefID"] != user_xrefid and
                    # # the message is not a MonitorStartResponse. This is the normal case where we send MonitorStart
                    # # at which time the user will not have xrefid assigned yet and we expect a response that we will
                    # # use to assign xrefid to the user
                    # inmessage_type != "MonitorStartResponse")
            ):

                # buffer received message if it is intended for another user or if it is an event that came sooner
                # than expected or if it is a response to a request from another user
                if (
                        # (this_user is not None and
                        #  inmessage["monitorCrossRefID"] and user_xrefid and
                        #  inmessage["monitorCrossRefID"] != user_xrefid and
                        #  inmessage["monitorCrossRefID"] in other_users_xrefid) or

                        (inmessage.is_event() and
                         inmessage["monitorCrossRefID"] and user_xrefid and
                         inmessage["monitorCrossRefID"] == user_xrefid) or

                        (inmessage.is_event() and
                         inmessage["monitorCrossRefID"] and
                         user_xrefid is None)
                        # or

                        # (inmessage.is_event() and
                        #  inmessage["callID"] in this_user.calls)

                        # (this_user is not None and
                        #  inmessage.is_response() and
                        #  {inmessage.eventid: inmessage_type.replace("Response", "")} in other_users_transactions)
                ):
                    with net_object.lock:
                    # with self.lock:
                        # trying global lock to buffer messages
                        self.buffer_message(net_object, inmessage, ntfy_subs=False)
                    # warning(
                    #     "BUFFERED MESSAGE '{}' with callID '{}' for '{}' because I am '{}' waiting for '{}' in '{}' '{}'".format(
                    #         inmessage_type,
                    #         inmessage["callID"],
                    #         inmessage["deviceID"], for_user, message,
                    #         this_user.monitorCrossRefID, this_user.callID))
                    inmessage = None
                else:
                    if this_user is None:
                        raise AssertionError('Got "{}" with callID {} and xrefid {} '
                                             'while expecting "{}"'.format(inmessage_type,
                                                                           inmessage["callID"],
                                                                           inmessage["monitorCrossRefID"],
                                                                           message))
                    else:
                        raise AssertionError(
                            'Got "{}" with callID {}, eventid {} and xrefid {} while expecting "{}" with '
                            'callID {} and xrefid {}. '
                            'Known transactions are:\n"{}"\n{} '.format(inmessage_type,
                                                                        inmessage["callID"],
                                                                        inmessage.eventid,
                                                                        inmessage["monitorCrossRefID"],
                                                                        message,
                                                                        this_user.parameters.get(
                                                                            "callID",
                                                                            None),
                                                                        this_user.parameters.get(
                                                                            "monitorCrossRefID",
                                                                            None),
                                                                        this_user.out_transactions,
                                                                        inmessage))
        if this_user is not None:
            # Evaluate the invoke id
            this_user.update_incoming_transactions(inmessage)
            this_user.update_call_id(inmessage)

        return inmessage

    def buffer_message(self, user, message, ntfy_subs=True):
        """
        Add csta message to csta user's or csta application's buffer.
        :param message: the message to buffer
        :param ntfy_subs: notify the user whose buffer was changed
        :param user: the user who's buffer to use
        :return: None
        """
        if user is self:
            warning("Unexpected message. Adding to global buffer: " + str(message))

        key = str(message["monitorCrossRefID"]) + str(message.event)
        if key in user.message_buffer:
            user.message_buffer[key].append(message)
        else:
            user.message_buffer[key] = [message]
        user.buffer_mod_time = time()
        if ntfy_subs:
            self.buffer_trigger_count += 1
        if self.buffer_trigger_count:
            user.buffer_event.set()
            self.buffer_trigger_count -= 1

    def update_call_parameters(self, directory_number, inresponse):
        """ Update our parameters based on the given incoming CSTA message """
        try:
            if not inresponse.event.endswith("Event") and not inresponse.event.endswith("Response"):
                self.parameters[directory_number]["last_request_eventid"] = inresponse.eventid
            for key in "deviceID", "monitorCrossRefID", "callID":
                if inresponse[key]:
                    self.parameters[directory_number][key] = inresponse[key]
        except AttributeError:
            self.parameters[directory_number].setdefault("deviceID", None)

    def close(self):
        self.link.socket.close()
        self.link = None



"""\
Purpose: CSTA message object
Initial Version: Costas Skarakis 11/11/2018
"""
import os
import re
import xml.etree.ElementTree as ET
import io
from common import util

xmlpath = os.path.join(os.path.realpath(__file__).replace(os.path.basename(__file__), ''), "CstaPool")


def is_response(message):
    if message.endswith("Response"):
        return True
    else:
        return False


def is_request(message):
    if is_event(message) or is_response(message):
        return False
    else:
        return True


def is_event(message):
    return message.endswith("Event")


class CstaMessage(object):
    """
    Representation of a CSTA message
    """

    def __init__(self, header, body):
        # s_ip,s_port,d_ip,d_port
        self.eventid = int(header[-4:])
        self.size = header[:4]
        self.header = header
        self.encoding = body.encoding
        self.body = body
        self.root = body.root
        self.namespace = body.namespace
        self.event = body.event

    def is_response(self):
        return is_response(self.event)

    def is_request(self):
        return is_request(self.event)

    def is_event(self):
        return is_event(self.event)

    def find_element(self, key):
        """
        Obsolete. Kept for backwards compatibility
        :param key:
        :return:
        """
        element = None
        for tag in self.body.iter():
            element = tag.find(key)
            if element is None:
                element = tag.find("{" + self.namespace + "}" + key)
            if element is None:
                element_search = re.search(key, tag.tag)
                if element_search:
                    element = tag
            if element is not None:
                break
        return element

    def __getitem__(self, key):
        element = self.body.get_tag(key)
        if element is None:
            return None
        else:
            return element.text

    def get_all(self, key):
        elements = self.body.get_all(key)
        if elements is None:
            return []
        else:
            return [element.text for element in elements]

    def __setitem__(self, key, value):
        element = self.body.get_tag(key)
        element.text = value

    def __repr__(self):
        return repr(self.body)

    def __str__(self):
        return repr(self)

    def message(self):
        return str(self)

    def set_eventid(self, eventid):
        self.header = self.header[:4] + bytes("%04d" % eventid, encoding=self.encoding)
        self.eventid = eventid

    def contents(self):
        """ returns byte string """
        # Recalculate length
        self.size = bytes.fromhex("%08X" % (len(self.message()) + 8))
        return self.size + \
               self.header[-4:] + \
               bytes(self.message(), encoding=self.encoding)


if __name__ == "__main__":
    print(is_response("GetAgentStateResponse"))
    print(is_request("GetAgentStateResponse"))


"""\
Purpose: Simulate a SIP phone/line appearance/user
Initial Version: Costas Skarakis
"""
from threading import Timer, Event

from _socket import timeout as sock_timeout
import traceback
from common.tc_logging import exception, warning
from sip.SipParser import parseBytes, buildMessage
import common.util as util
import common.client as client
import sip.SipFlows as flow
from threading import Lock, Thread
from sip.SipMessage import SipMessage
from time import time, sleep


def dialog_hash(dialog):
    return "|".join([dialog["Call-ID"], dialog["from_tag"], dialog["to_tag"]])


class SipEndpoint(object):
    """\
    Representation of a SIP Endpoint
    """

    def __init__(self, directory_number):
        """
        Set some initial attribute values
        """
        self.ip = None
        self.port = None
        self.link = None
        self.number = directory_number
        self.parameters = {"user": directory_number,
                           "callId": None,
                           "fromTag": None,
                           "viaBranch": None,
                           "epid": None,
                           "expires": 360,
                           "cseq": "0",
                           "method": None
                           }
        self.last_messages_per_dialog = []
        self.dialogs = []
        self.known_call_ids = set()
        self.requests = []
        self.current_dialog = {
            "Call-ID": None,
            "from_tag": None,
            "to_tag": None
            # "epid": None,
        }
        self.tags = {}
        self.current_transaction = {"via_branch": None,
                                    "cseq": "0",
                                    "method": None
                                    }

        self.waitForMessage = self.wait_for_message  # compatibility alias
        self.reply_to = self.send_in_ctx_of  # method alias
        self.secondary_lines = []
        self.message_buffer = set()
        self.buffer_event = Event()
        # self.buffer_mod_count = [0]  # using a int in a list to be able to share and muate between objects in use_link()
        self.registered = False
        self.re_register_timer = None
        self.busy = False
        self.lock = Lock()
        self.wait_thread = None
        self.shutdown_flag = False

    def shutdown(self):
        """
        Try to stop threads and cleanup connections
        """
        self.shutdown_flag = True
        self.link.shutdown()
        self.wait_thread.join()

    def make_busy(self, busy=True):
        self.busy = busy

    def update_parameters(self, params, force=False):
        """
        Update endpoint parameters. This is useful for adding more flexibility in the creation of SIP messages

        :param params: The key-value pairs input
        :param force: Force update of existing variables
        :return: None
        """
        for parameter in params:
            if parameter not in self.parameters or force is True:
                self.parameters[parameter] = params[parameter]

    def connect(self, local_address, destination_address, protocol="tcp", certificate=None, subject_name="localhost"):
        """ Connect to the SIP Server """
        local_ip, local_port = local_address
        dest_ip, dest_port = destination_address

        self.parameters["dest_ip"] = dest_ip
        self.parameters["dest_port"] = dest_port
        self.parameters["transport"] = protocol
        if protocol in ("tcp", "TCP"):
            self.link = client.TCPClient(local_ip, local_port)
        elif protocol in ("udp", "UDP"):
            self.link = client.UDPClient(local_ip, local_port)
        elif protocol in ("tls", "TLS"):
            # for MTLS this might be needed
            # context.load_cert_chain('/path/to/certchain.pem', '/path/to/private.key')
            self.link = client.TLSClient(local_ip, local_port, certificate)
        else:
            raise NotImplementedError("{} client not implemented".format(protocol))
        self.link.connect(dest_ip, dest_port)
        self.set_address((local_ip, self.link.port))
        self.start_wait_thread()

    def start_wait_thread(self):
        """
        Starts a separate thread that will consume incoming csta traffic and place it into buffers
        """
        if self.wait_thread is None:
            self.wait_thread = Thread(target=self.wait_loop, daemon=True)
            self.wait_thread.start()

    def wait_loop(self):
        """
        Continuously wait for incoming sip messages. Add all messages to buffers
        """
        while not self.shutdown_flag:
            try:
                inbytes = self.link.waitForSipData(timeout=None)
                # inbytes = self.link.waitForSipData(timeout=timeout, client=link)
            except ConnectionError:
                # might need to add what kind of disconnection that was
                warning("Disconnected. Exiting")
                break
            if inbytes is None:
                warning("Disconnected. Will retry in 1 second")
                sleep(1)
            else:
                inmessage = parseBytes(inbytes)
                # TODO: Add "on message" functionality, eg 200OK on OPTIONS
                try:
                    self.message_buffer.add(inmessage)
                    self.buffer_event.set()
                except:
                    exception(traceback.format_exc())

    def update_to_tag(self, in_dialog):
        """
        Update the to_tag on an existing dialog that has no to_tag
        :param in_dialog: the complete dialog
        :return: None
        """
        with self.lock:
            for dialog in self.dialogs:
                if dialog["Call-ID"] == in_dialog["Call-ID"] \
                        and dialog["from_tag"] == in_dialog["from_tag"] \
                        and not dialog["to_tag"]:
                    dhash = dialog_hash(dialog)
                    dialog["to_tag"] = in_dialog["to_tag"]
                    dhash_complete = dialog_hash(dialog)
                    self.tags[dhash_complete] = self.tags[dhash]

    def get_complete_dialog(self, in_dialog):
        """
        Look for a complete existing dialog in case the provided one doesn't have a to tag
        :param in_dialog: the possibly incomplete dialog
        :return: in_dialog if it is complete or a corresponding complete one is not found,
                 otherwise the existing corresponding complete dialog
        """
        if in_dialog["to_tag"]:
            return in_dialog
        with self.lock:
            for dialog in self.dialogs:
                if dialog["Call-ID"] == in_dialog["Call-ID"] \
                        and dialog["from_tag"] == in_dialog["from_tag"] \
                        and dialog["to_tag"]:
                    return dialog
            return in_dialog

    def set_address(self, address):
        """
        We can create and Endpoint just to use it as target for sip messages

        :param address: The address as an (ip, port) tuple
        :return: None
        """
        local_ip, local_port = address
        self.ip = local_ip
        self.port = local_port
        self.parameters["source_ip"] = local_ip
        self.parameters["source_port"] = local_port

    def use_link(self, other):
        """ Convenience function to use an existing network connection
        If parameter is SipEndpoint, we will use also share a common message buffer and dialog tracking
        """
        if isinstance(other, SipEndpoint):
            link = other.link
            self.message_buffer = other.message_buffer
            self.dialogs = other.dialogs
            self.requests = other.requests
            self.last_messages_per_dialog = other.last_messages_per_dialog
            self.buffer_event = other.buffer_event
            self.wait_thread = other.wait_thread
        else:
            link = other
        protocol = ["TCP", "UDP"][link.socket.proto]
        local_ip = link.ip
        local_port = link.port
        dest_ip = link.rip
        dest_port = link.rport
        self.link = link
        self.ip = local_ip
        self.port = local_port
        self.parameters["source_ip"] = local_ip
        self.parameters["source_port"] = local_port
        self.parameters["dest_ip"] = dest_ip
        self.parameters["dest_port"] = dest_port
        self.parameters["transport"] = protocol
        # self.link.endpoints_connected += 1

    def get_dialog(self):
        """ Will use this method to get thread-local dialogs """
        if "toTag" not in self.parameters:
            return {"Call-ID": self.parameters["callId"],
                    "from_tag": self.parameters["fromTag"]}
        else:
            return {"Call-ID": self.parameters["callId"],
                    "from_tag": self.parameters["fromTag"],
                    "to_tag": self.parameters["toTag"]}

    def get_transaction(self):
        """ Will use this method to get thread-local transactions """
        return {"via_branch": self.parameters["viaBranch"],
                "cseq": self.parameters["cseq"],
                "method": self.parameters["method"]}

    def set_dialog(self, dialog):
        """ Change current dialog to the one provided """
        if not isinstance(dialog, dict):
            exception("Must provide a dialog in the form of Python dictionary")
        for key in self.get_dialog():
            if key not in dialog:
                exception("Not a valid dialog. Missing key: " + key)
        dialog = self.get_complete_dialog(dialog)
        if dialog not in self.dialogs:
            self.dialogs.append(dialog)
            self.known_call_ids.add(dialog["Call-ID"])
            self.requests.append([])
            # If we don't know this dialog it means we didn't started so it must be an incoming Request
            with self.lock:
                self.tags[dialog_hash(dialog)] = "to_tag"
        self.current_dialog = dialog
        self.parameters["callId"] = dialog["Call-ID"]
        self.parameters["fromTag"] = dialog["from_tag"]
        self.parameters["toTag"] = dialog["to_tag"]
        return dialog

    def reset_dialog_and_transaction(self):
        """
        Reset current dialog and transaction to initial state
        :return: None
        """
        self.parameters["viaBranch"] = None
        self.parameters["epid"] = None
        self.parameters["cseq"] = "0"
        self.parameters["method"] = None
        self.parameters["callId"] = None
        self.parameters["fromTag"] = None
        self.parameters["toTag"] = None
        self.current_dialog = {
            "Call-ID": None,
            "from_tag": None,
            "to_tag": None
        }
        self.current_transaction = {"via_branch": None,
                                    "cseq": "0",
                                    "method": None
                                    }

    def switch_tags(self, dialog=None):
        if not dialog:
            dialog = self.get_dialog()
        dhash = dialog_hash(dialog)
        dialog["from_tag"], dialog["to_tag"] = dialog["to_tag"], dialog["from_tag"]
        dhash_switched = dialog_hash(dialog)
        self.set_dialog(dialog)
        with self.lock:
            if self.tags[dhash] == "to_tag":
                self.tags[dhash_switched] = "from_tag"
            else:
                self.tags[dhash_switched] = "to_tag"

    def set_transaction(self, transaction):
        """ Change current transaction to the one provided """
        if not isinstance(transaction, dict):
            exception("Must provide a transaction in the form of Python dictionary")
            return
        for key in self.current_transaction:
            if key not in transaction:
                exception("Not a valid transaction. Missing key: " + key)
                return
        for key in self.current_transaction:
            self.current_transaction[key] = transaction[key]
        self.parameters["viaBranch"] = transaction["via_branch"]
        self.parameters["cseq"] = transaction["cseq"]
        self.parameters["method"] = transaction["method"]

    def start_new_dialog(self):
        """ Refresh the SIP dialog specific parameters """
        dialog = {
            "Call-ID": util.randomCallID(),
            "from_tag": util.randomTag(),
            "to_tag": ""
            # "epid": lambda x=6: "SC" + util.randStr(x),
        }
        self.current_dialog = dialog
        self.parameters["callId"] = dialog["Call-ID"]
        self.parameters["fromTag"] = dialog["from_tag"]
        self.parameters["toTag"] = dialog["to_tag"]
        with self.lock:
            self.tags[dialog_hash(dialog)] = "from_tag"
        self.dialogs.append(dialog)
        self.known_call_ids.add(dialog["Call-ID"])
        self.requests.append([])
        return dialog

    def get_last_message_in(self, dialog):
        """ Get the last message sent or received in the provided dialog """
        # If we have received no messages yet return None
        null_d = {"Call-ID": None, "from_tag": None, "to_tag": None}
        if self.current_dialog == null_d or dialog == null_d:
            return None
        with self.lock:
            # First check for complete dialogs
            for message in self.last_messages_per_dialog:
                if message.get_dialog() == dialog:
                    return message
            # Second check for incomplete dialogs
            for message in self.last_messages_per_dialog:
                d = message.get_dialog()
                if d["Call-ID"] == dialog["Call-ID"] and d["from_tag"] == dialog["from_tag"]:
                    return message
        raise Exception("No message found in dialog {}. Other dialogs active: ".format(dialog, self.dialogs))

    def save_message(self, message):
        """ Search for previously received message in the same dialog.
            If found, replace with given message, otherwise append message to message list """
        with self.lock:
            for i in range(len(self.last_messages_per_dialog)):
                if message.get_dialog() == self.last_messages_per_dialog[i].get_dialog():
                    self.last_messages_per_dialog[i] = message
                    return
            self.last_messages_per_dialog.append(message)

    def prepare_message(self, message):
        """
        Convert a SIP message string to SipMessage object

        :param message: The message string
        :return: A SipMessage object
        """
        return buildMessage(message, self.parameters)

    def start_new_transaction(self, method, dialog=None):
        """ Refresh the via branch and CSeq header """
        if not dialog:
            dialog = self.get_dialog()
        try:
            last_message_in_dialog = self.get_last_message_in(dialog)
        except:
            last_message_in_dialog = None
        branch = util.randomBranch()
        if method in ("ACK", "CANCEL"):
            # Not really a new transaction
            # find transaction from last message in dialog
            # which should not be none because what are we sending ACK or CANCEL to?
            transaction = last_message_in_dialog.get_transaction()
            cseq = transaction["cseq"]
        elif last_message_in_dialog:
            transaction = last_message_in_dialog.get_transaction()
            cseq = str(int(transaction["cseq"]) + 1)
        else:
            with self.lock:
                cseq = str(len(self.requests[self.dialogs.index(dialog)]))
                if method != "REGISTER":
                    # Ignore REGISTER otherwise unregister breaks
                    # TODO: check if reRegistrations will work
                    self.requests[self.dialogs.index(dialog)].append(method)
        transaction = {"via_branch": branch, "cseq": cseq, "method": method}
        self.set_transaction(transaction)
        return transaction

    def send_new(self, target_sip_ep=None, message_string="", expected_response=None, ignore_messages=[]):
        """ Start a new dialog and send a message """
        self.parameters["userA"] = self.number
        if target_sip_ep:
            if isinstance(target_sip_ep, SipEndpoint):
                self.parameters["userB"] = target_sip_ep.number
                target_sip_ep.parameters["userB"] = self.number
            elif isinstance(target_sip_ep, str):
                self.parameters["userB"] = target_sip_ep
            else:
                raise Exception("target_sip_ep must be str or SipEndpoint, not {}".format(type(target_sip_ep)))
        else:
            # In cases like REGISTER, there is no B-side
            # Clear parameters to avoid unexpected results
            # self.parameters.pop("userA", None)
            # self.parameters.pop("userB", None)
            pass

        if isinstance(message_string, SipMessage):
            m = message_string
        else:
            m = buildMessage(message_string, self.parameters)
        assert m.type == "Request", 'Tried to start a new dialog with a SIP Response'

        new_dialog = self.start_new_dialog()
        new_transaction = self.start_new_transaction(m.method)  # m should always be a request

        m.set_dialog_from(new_dialog)
        m.set_transaction_from(new_transaction)

        self.link.send(m.contents())
        self.save_message(m)

        if expected_response:
            # try:
            self.waitForMessage(message_type=expected_response, dialog=new_dialog, ignore_messages=ignore_messages)
        # except AssertionError:
        #     raise AssertionError('{}: "{}" response to "{}"\n{}'.format(self.number,
        #                                                                 self.get_last_message_in(
        #                                                                     m.get_dialog()).get_status_or_method(),
        #                                                                 m.method,
        #                                                                 m))
        return m

    def send_in_ctx_of(self, reference_message, this_message_string="", expected_response=None, ignore_messages=[]):
        """ Send a message within the same dialog and transaction as 'reference_message' """
        self.set_transaction(reference_message.get_transaction())
        return self.send(message_string=this_message_string,
                         expected_response=expected_response,
                         ignore_messages=ignore_messages,
                         dialog=reference_message.get_dialog())

    def send(self, message_string="", expected_response=None, ignore_messages=[], dialog=None):
        """ Send a message within a dialog """
        m = self.reply(message_string, dialog)
        if expected_response:
            self.waitForMessage(message_type=expected_response, ignore_messages=ignore_messages, dialog=dialog)
        return m

    def reply(self, message_string, dialog=None):
        """ Send a response to a previously received message """
        if "callId" not in self.parameters or not self.parameters['callId']:
            raise Exception("Cannot reply when we are not in a dialog")

        if isinstance(message_string, SipMessage):
            # We can also get a SipEndpoint type as input but it will be sent as is
            m = message_string
        else:
            m = buildMessage(message_string, self.parameters)

        if dialog:
            dialog = self.set_dialog(dialog)
        else:
            dialog = self.get_dialog()

        try:
            previous_message = self.get_last_message_in(dialog)
            m.make_response_to(previous_message)
            self.update_to_tag(m.get_dialog())
        except:  # caused by get_last_message_in(dialog) if no message has been exchanged in this dialog yet
            # New dialog, same call-id, eg NOTIFY after Keyset SUBSCRIBE.
            # Must be a SIP Request
            previous_message = None
            assert m.type == "Request", \
                "Attempted to send a {} response in a new dialog".format(m.get_status_or_method())

        if m.type == "Request":
            # This is a new request in the same dialog, so fix the CSeq
            # m.increase_cseq()
            self.start_new_transaction(m.method)
            # If B side sends the new request we must switch from and to tags
            if previous_message is not None and \
                    m.method not in ("ACK", "CANCEL") and \
                    self.tags[dialog_hash(dialog)] == "to_tag":
                self.switch_tags(dialog)
            m.set_dialog_from(dialog)
        else:
            self.free_resources(m)
        self.save_message(m)

        m.set_transaction_from(self.get_transaction())
        self.link.send(m.contents())
        return m

    def get_buffered_message(self, message_type, dialog):
        """
        Return the first buffered message found in the given dialog

        :param message_type: the requested message type
        :param dialog: The dialog in question
        :return: the buffered SipMessage
        """
        msg = None
        # for message in self.message_buffer:
        with self.lock:
            for message in self.message_buffer:
                m_dialog = message.get_dialog()
                m_type = message.get_status_or_method()
                # If we have received no messages yet return the first message in the buffer
                if message_type in m_type and (dialog is None or m_dialog["Call-ID"] == dialog["Call-ID"]):
                    msg = message
                    self.message_buffer.remove(message)
                    break
                # else:
                #     self.message_buffer.append(message)
        # if msg is None:
        #     print(self.number, "Buffer returned None for", message_type, "in callid and from_tag", dialog["Call-ID"], dialog["from_tag"], "although it has", self.message_buffer)
        return msg

    def wait_for_message(self, message_type, dialog=None, ignore_messages=(), timeout=5.0):
        """
        Wait for a specific type of SIP message.
        :param message_type: is a string that we will make sure is
                             contained in the received message request or response line
                             Set this to None or "" to accept any incoming message

                             if a list or tuple is given any of the contained message types
                             will be accepted

        :param dialog: Set the dialog to expect the message in. If None will expect a message in current dialog
        :param ignore_messages: a list of message_types to ignore if they come before
                                the expected message_type
        :param timeout: Defined timeout in seconds.
        :return: A SipMessage constructed from the incoming message
        """

        # if not dialog:
        #     dialog = {"Call-ID": self.parameters["callId"],
        #               "from_tag": self.parameters["fromTag"]}
        #     if "toTag" in self.parameters:
        #         dialog["to_tag"] = self.parameters["toTag"]

        if dialog:
            dialog = self.get_complete_dialog(dialog)
            explicit_dialog = dialog
        else:
            explicit_dialog = None
            dialog = {"Call-ID": self.parameters["callId"],
                      "from_tag": self.parameters["fromTag"]}
            if "toTag" in self.parameters:
                dialog["to_tag"] = self.parameters["toTag"]

        last_sent_message = self.get_last_message_in(dialog)
        transaction = None
        if last_sent_message:
            transaction = last_sent_message.get_transaction()

        t0_tout = time()
        inmessage = self.get_buffered_message(message_type, explicit_dialog)
        inmessage = self.handleDA(last_sent_message, inmessage)

        while not inmessage:
            rem_timeout = timeout - (time() - t0_tout)
            if rem_timeout > 0:
                self.buffer_event.wait(rem_timeout)
                self.buffer_event.clear()
                inmessage = self.get_buffered_message(message_type, explicit_dialog)
                inmessage = self.handleDA(last_sent_message, inmessage)
                if inmessage is None:
                    continue
            else:
                exception("%s No %s. Buffer length: %d." % (self.number,
                                                            message_type,
                                                            len(self.message_buffer)))
                raise sock_timeout

            # if not inmessage:
            #     # no (more) buffered messages. try the network
            #     inbytes = self.link.waitForSipData(timeout=timeout, client=link)
            #     inmessage = self.handleDA(last_sent_message, parseBytes(inbytes))

            inmessage_type = inmessage.get_status_or_method()
            inmessage_dialog = inmessage.get_dialog()
            inmessage.cseq_method = inmessage.get_transaction()["method"]

            # when sharing buffers we can get a message of other endpoints
            # buffer message of correct type but unknown callid
            # keep it if we are mentioned in the "To" header
            if inmessage_type == message_type and \
                    inmessage_dialog["Call-ID"] not in self.known_call_ids and \
                    "sip:{}@{}".format(self.number, self.ip) not in inmessage["To"]:
                # print(self.number, "Aborting", inmessage_type, "with callid", inmessage_dialog["Call-ID"])
                # print(self.number, "My callid is", dialog["Call-ID"])
                with self.lock:
                    self.message_buffer.add(inmessage)
                self.buffer_event.set()
                inmessage = None
                continue

            if inmessage_type in ignore_messages:
                inmessage = None
                continue

            if message_type and \
                    ((isinstance(message_type, str) and message_type not in inmessage_type) or
                     (type(message_type) in (list, tuple) and not any([m in inmessage_type for m in message_type])) or
                     (inmessage.type == "Response" and inmessage.cseq_method != transaction["method"])):
                # we have received an unexpected message. buffer it if there is an active dialog for it
                if self.get_complete_dialog(inmessage_dialog) or inmessage_type == "INVITE":
                    # message is part of another active dialog or a new call, so buffer it
                    # print(self.number, "Aborting", inmessage_type, "with callid", inmessage_dialog["Call-ID"])
                    with self.lock:
                        self.message_buffer.add(inmessage)
                    self.buffer_event.set()
                    # print("Appended {} with {} to buffer. Will keep waiting for {} in {} ".format(inmessage_type,
                    #                                                                        inmessage_dialog,
                    #                                                                        message_type,
                    #                                                                            dialog))
                    inmessage = None
                else:
                    d = ["sip:{}@".format(line.number) in inmessage["To"] for line in self.secondary_lines]
                    if any(d):
                        # message is meant for another line in this device
                        with self.lock:
                            self.secondary_lines[d.index(True)].message_buffer.add(inmessage)
                        inmessage = None
                    else:
                        raise AssertionError('{}: Got "{}" in {} while expecting "{}" in {}. '
                                             'Other active dialogs:{}.'.format(self.number,
                                                                               inmessage_type,
                                                                               inmessage_dialog,
                                                                               message_type,
                                                                               dialog,
                                                                               self.dialogs))

        self.save_message(inmessage)
        if inmessage.type == "Request":
            inmessage_transaction = inmessage.get_transaction()
            self.set_transaction(inmessage_transaction)
        else:
            assert inmessage.get_transaction()["method"] == transaction["method"], \
                "Got {} to {} instead of {} to {}".format(inmessage.get_status_or_method(),
                                                          inmessage.get_transaction()["method"],
                                                          message_type,
                                                          transaction["method"])
            self.update_to_tag(inmessage.get_dialog())
            self.free_resources(inmessage)
        self.set_dialog(inmessage.get_dialog())
        return inmessage

    def wait_for_messages(self, *list_of_message_types, in_order=False, ignore_messages=[]):
        """
        Wait for a list of messages. To be used in a loop or with the next() function.
        Each invocation will return the next SipMessage received.
        The loop will exit when all messages have been received

        :param list_of_message_types: The list of messages expected eg ["NOTIFY", "SUBSCRIBE", "403 Forbidden"]
        :param in_order: Impose the given order for incoming messages
        :param ignore_messages:
        :return: SipMessage received after each invocation
        """
        l = len(list_of_message_types)
        not_received_messages = list(list_of_message_types)
        received_messages = []
        for i in range(l):
            try:
                sip_message = self.wait_for_message(None, ignore_messages)
            except TimeoutError:
                raise TimeoutError("Timeout waiting for {}. Messages already received: {}".format(not_received_messages,
                                                                                                  received_messages))
            message_received = sip_message.get_status_or_method()
            assert message_received in not_received_messages, \
                'After receiving {} we expected one of {} but got {}'.format(received_messages,
                                                                             not_received_messages,
                                                                             message_received)
            if in_order:
                message_order = list_of_message_types.index(message_received)
                assert i == message_order, \
                    '{} arrived in order {} instead of {}'.format(message_received, i, message_order)
            not_received_messages.remove(message_received)
            received_messages.append(message_received)
            yield sip_message

    def set_digest_credentials(self, username, password, realm=""):
        """
        Set the digest authentication credentials
        :param username: digest username
        :param password: digest password
        :param realm: digest realm - no effect at the moment since it is overridden by 401 response realm
        """
        self.parameters.update({
            "da_user": username,
            "da_pass": password,
            "realm": realm
        })

    def handleDA(self, request, response):
        """" Add DA to message and send again """
        if response is None:
            return
        # Usual case in lab, password same as username
        if "da_pass" not in self.parameters or "da_user" not in self.parameters:
            self.set_digest_credentials(self.number, self.number, "")
        user, pwd = self.parameters["da_user"], self.parameters["da_pass"]
        if response.type == "Response" and response.status == "401 Unauthorized":
            request.addAuthorization(response["WWW-Authenticate"], user, pwd)
            self.link.send(request.contents())
            inBytes = self.link.waitForSipData()
            return parseBytes(inBytes)
        else:
            return response

    def register(self, expiration_in_seconds=360, re_register_time=180):
        """ Convenience function to register a SipEndpoint """
        if not expiration_in_seconds:
            return self.unregister()
        if re_register_time and self.registered:
            self.re_register_timer = Timer(re_register_time, self.register, (expiration_in_seconds, re_register_time))
            self.re_register_timer.start()
        flow.register(self, expiration_in_seconds)
        self.reset_dialog_and_transaction()
        self.registered = True

    def unregister(self):
        """ Convenience function to un-register a SipEndpoint"""
        if self.re_register_timer:
            self.re_register_timer.cancel()
        flow.unregister(self)
        self.registered = False

    def free_resources(self, message):
        """ Free memory allocated to a cleared call """

        status_or_method = message.get_status_or_method()
        cseq_method = message["CSeq"].split()[1]
        call_ended_successfully = status_or_method.startswith("2") and cseq_method in ("BYE", "CANCEL")
        call_rejected_with_error = status_or_method[0] in "3456"
        if call_ended_successfully or call_rejected_with_error:
            dialog = message.get_dialog()
            with self.lock:
                # TODO: actually free up memory from completed calls
                dialog_index = self.dialogs.index(dialog)
                # self.dialogs.pop(dialog_index)
                # self.requests.pop(dialog_index)
                # self.tags.pop(dialog_hash(dialog))
                # self.last_messages_per_dialog.pop(dialog_index)
                # print("Cleared call number", dialog_index, "after message", status_or_method, "with cseq", cseq_method,
                #       ". Sizes are", len(self.dialogs), len(self.requests), len(self.tags))



"""\
Purpose: Functions for parsing CSTA messages and creating CstaMessage objects
Initial Version: Costas Skarakis 11/11/2018
"""
import re
import xml.etree.ElementTree as ET
from csta.CstaMessage import CstaMessage
from common.util import XmlBody

ENCODING = "utf8"


def buildMessageFromFile(filename, parameters, eventid):
    with open(filename, "r") as file:
        return buildMessage(file.read(), parameters, eventid)


def buildMessage(message, parameters={}, eventid=0):
    encoding_search = re.search("encoding=[\'\"](\S*)[\'\"].* ?\?\>", message)
    if not encoding_search:
        print("Warning: No encoding defined in message:", message)
        xml_encoding = ENCODING
    else:
        xml_encoding = encoding_search.group(1)
    tString = message.strip().format(**parameters)
    # bString=bytes(tString.replace("\n","\r\n")+2*"\r\n",encoding=xml_encoding)
    bString = bytes("%04d" % eventid + tString, encoding=xml_encoding)
    bHeader = bytes.fromhex("%08X" % (len(bString) + 4))
    cstaMessage = parseBytes(bHeader + bString)
    return cstaMessage


def parseBytes(bString):
    try:
        xml_encoding = re.search(b"encoding=[\'\"](\S*)[\'\"].* ?\?\>", bString).group(1)
        encoding = xml_encoding.decode(encoding=ENCODING)
    except AttributeError:
        encoding = ENCODING
    header = bString[:8]
    try:
        body = XmlBody(bString[8:].strip().decode(encoding))
    except:
        print("Cannot decode CSTA message with ", encoding)
        print(bString)
        raise
    try:
        root = body.root
    except:
        print("Cannot parse CSTA message string:")
        print(header)
        print(body)
        raise

    return CstaMessage(header, body)


if __name__ == "__main__":
    #    import os
    #    os.chdir(r'.\CstaPool')
    #    s=buildMessageFromFile("SystemRegister.xml",{"deviceID":10101001100,"reason":"because"},9999)
    a = []
    b = []
    a.append(b'\x00\x00\x03\xeb9999<?xml version="1.0" encoding="UTF-8"?><ServiceInitiatedEvent xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><monitorCrossRefID>1169730059</monitorCrossRefID><initiatedConnection><callID>FF000100000000002D65C85C1C750000</callID><deviceID>302118840100</deviceID></initiatedConnection><initiatingDevice><deviceIdentifier>N&lt;302118840100&gt;;noa=nk</deviceIdentifier></initiatingDevice><localConnectionInfo>initiated</localConnectionInfo><cause>makeCall</cause><servicesPermitted><callControlServices><clearConnection>true</clearConnection></callControlServices><callAssociatedServices/><mediaAttachementServices/><routeingServices/><voiceServices/></servicesPermitted><mediaCallCharacteristics><mediaClass><voice>true</voice><image>false</image><im>false</im></mediaClass></mediaCallCharacteristics><extensions><privateData><private xmlns:scx="http://www.siemens.com/schema/csta"><scx:extendedServicesPermitted/></private></privateData></extensions></ServiceInitiatedEvent>')
    a.append(b'\x00\x00\x04C9999<?xml version="1.0" encoding="UTF-8"?><OriginatedEvent xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><monitorCrossRefID>1169730059</monitorCrossRefID><originatedConnection><callID>FF000100000000002D65C85C1C750000</callID><deviceID>302118840100</deviceID></originatedConnection><callingDevice><deviceIdentifier>302118840100</deviceIdentifier></callingDevice><calledDevice><deviceIdentifier>N&lt;302118840101&gt;;displayNumber=302118840101;noa=nk</deviceIdentifier></calledDevice><localConnectionInfo>connected</localConnectionInfo><cause>newCall</cause><servicesPermitted><callControlServices><clearConnection>true</clearConnection></callControlServices><callAssociatedServices/><mediaAttachementServices/><routeingServices/><voiceServices/></servicesPermitted><mediaCallCharacteristics><mediaClass><voice>true</voice><image>false</image><im>false</im></mediaClass></mediaCallCharacteristics><extensions><privateData><private xmlns:scx="http://www.siemens.com/schema/csta"><scx:extendedServicesPermitted/></private></privateData></extensions></OriginatedEvent>')
    a.append(b'\x00\x00\x04A0002<?xml version="1.0" encoding="UTF-8"?><MonitorStop xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><monitorCrossRefID>1193240031</monitorCrossRefID><extensions><privateData><private xmlns:scx="http://www.siemens.com/schema/csta"><scx:staticOndDN>17133013</scx:staticOndDN></private></privateData></extensions></MonitorStop>')
    a.append(b'\x00\x00\xb4A0002<?xml version="1.0" encoding="UTF-8"?><StartApplicationSessionPosResponse xmlns:aps="http://www.ecma-international.org/standards/ecma-354/appl_session"><sessionID>009968951</sessionID><actualProtocolVersion>http://www.ecma-international.org/standards/ecma-323/csta/ed4</actualProtocolVersion><actualSessionDuration>600</actualSessionDuration><extensions><actualHBT>0</actualHBT></extensions></StartApplicationSessionPosResponse>')
    b.append('''<?xml version="1.0" encoding="utf-8"?>
<MonitorStart>
   <monitorObject>
      <deviceObject>{user}</deviceObject>
   </monitorObject>
   <monitorType>device</monitorType>
   <extensions>
      <privateData>
         <private xmlns:scx="http://www.siemens.com/schema/csta">
            <scx:staticOndDN>17133013</scx:staticOndDN>
         </private>
      </privateData>
   </extensions>
</MonitorStart>''')
    a.append(b'\x00\x00\x04A0002<?xml version="1.0" encoding="UTF-8"?><SystemRegisterResponse xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><sysStatRegisterID>27ss</sysStatRegisterID></SystemRegisterResponse>')
    a.append(b'\x00\x00\x04A0002<?xml version="1.0" encoding="UTF-8"?><MonitorStartResponse xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><monitorCrossRefID>1136370025</monitorCrossRefID></MonitorStartResponse>')
    b.append('''<?xml version="1.0" encoding="utf-8"?>
<StartApplicationSession>
   <applicationInfo>
      <applicationID>pythontsm</applicationID>
   </applicationInfo>
   <requestedProtocolVersions>
      <protocolVersion>http://www.ecma-international.org/standards/ecma-323/csta/ed4</protocolVersion>
   </requestedProtocolVersions>
   <requestedSessionDuration>600</requestedSessionDuration>
</StartApplicationSession>''')
    b.append('''<?xml version="1.0" encoding="UTF-8"?>
<SystemStatus
  xmlns="http://www.ecma.ch/standards/ecma-323/csta/ed2"
  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
  xsi:schemaLocation="
      http://www.ecma.ch/standards/ecma-323/csta/ed2  file://localhost/X:/ips_bln/long_csta/ecma/system-status.xsd
  ">
  <systemStatus>normal</systemStatus>
</SystemStatus>''')
    b.append('''<?xml version="1.0" encoding="UTF-8" standalone="yes"?><ns3:SystemStatus xmlns:ns2="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns3="http://www.ecma-international.org/standards/ecma-323/csta/ed4"/>''')
    b.append('''<?xml version="1.0" encoding="UTF-8" standalone="yes"?><ns3:MonitorStart xmlns:ns2="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns3="http://www.ecma-international.org/standards/ecma-323/csta/ed4"><ns3:monitorObject><ns3:deviceObject typeOfNumber="dialingNumber">3021033</ns3:deviceObject></ns3:monitorObject></ns3:MonitorStart>''')
    b.append('<?xml version="1.0" encoding="utf-8"?><DeflectCall><callToBeDiverted><deviceID>SUB_C</deviceID><callID>FF000100000000006BB1234C31000000</callID></callToBeDiverted><newDestination>SUB_D</newDestination></DeflectCall>')
    b.append('''<?xml version='1.0' encoding='utf-8'?><CallInformationEvent xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4" xmlns:scx="http://www.siemens.com/schema/csta"><monitorCrossRefID>1166743342</monitorCrossRefID><connection><callID>FF00010000000000E2E00E6416060000</callID><deviceID>3021061290000</deviceID></connection><device><deviceIdentifier>N&lt;3021061290000&gt;;noa=nk</deviceIdentifier></device><correlatorData><string>75726E3A6E656E613A7569643A63616C6C69643A7535543341316E627635462E6F7376636932306E312E68386B2E7365633B75726E3A6E656E613A7569643A696E636964656E7469643A527A6E447039543643756E2E6F7376636932306E312E68386B2E736563</string></correlatorData><connectionInfo><mediaSessionInfo>voice</mediaSessionInfo></connectionInfo><callCharacteristics><acdCall>true</acdCall><priorityCall>true</priorityCall><highPriorityCall>true</highPriorityCall><sensitiveCall>true</sensitiveCall></callCharacteristics><extensions><privateData><private><scx:aliQueryTimeout><scx:aliDBInfo>10.2.31.5:9991</scx:aliDBInfo><scx:aliConfiguredTimeout>200</scx:aliConfiguredTimeout></scx:aliQueryTimeout></private></privateData></extensions></CallInformationEvent>''')
    b.append('''<?xml version="1.0" encoding="UTF-8"?>
<DeliveredEvent xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4">
   <monitorCrossRefID>1173673890</monitorCrossRefID>
   <connection>
      <callID>FF00010000000000EC6114647E570000</callID>
      <deviceID>3021061290002</deviceID>
   </connection>
   <alertingDevice>
      <deviceIdentifier>N&lt;3021061290002&gt;;noa=nk</deviceIdentifier>
   </alertingDevice>
   <callingDevice>
      <deviceIdentifier>N&lt;7868772298&gt;Line71001;displayNumber=7868772298;uid=FF00010000000000EC6114647E570000</deviceIdentifier>
   </callingDevice>
   <calledDevice>
      <deviceIdentifier>N&lt;3021061290002&gt;;noa=nk</deviceIdentifier>
   </calledDevice>
   <lastRedirectionDevice>
      <notRequired/>
   </lastRedirectionDevice>
   <localConnectionInfo>alerting</localConnectionInfo>
   <correlatorData>
      <string>
         7
         5
         7
         2
         6
         E
         3
         A6E656E613A7569643A63616C6C69643A42544F43696C6A7332344D2E6F7376636932306E312E68386B2E7365633B75726E3A6E656E613A7569643A696E636964656E7469643A746669477346763462525A2E6F7376636932306E312E68386B2E736563
      </string>
   </correlatorData>
   <cause>enteringDistribution</cause>
   <servicesPermitted>
      <callControlServices>
         <clearConnection>true</clearConnection>
         <deflectCall>true</deflectCall>
      </callControlServices>
      <callAssociatedServices/>
      <mediaAttachementServices/>
      <routeingServices/>
      <voiceServices/>
   </servicesPermitted>
   <networkCalledDevice>
      <deviceIdentifier>02911</deviceIdentifier>
   </networkCalledDevice>
   <associatedCallingDevice>
      <notKnown/>
   </associatedCallingDevice>
   <mediaCallCharacteristics>
      <mediaClass>
         <voice>true</voice>
         <image>false</image>
         <message>false</message>
      </mediaClass>
   </mediaCallCharacteristics>
   <callCharacteristics>
      <acdCall>true</acdCall>
      <priorityCall>true</priorityCall>
      <highPriorityCall>true</highPriorityCall>
      <sensitiveCall>true</sensitiveCall>
   </callCharacteristics>
   <extensions>
      <privateData>
         <private xmlns:scx="http://www.siemens.com/schema/csta">
            <scx:cdrTrunkSeizureDateTime>2023/03/17 14:49:48</scx:cdrTrunkSeizureDateTime>
            <scx:extendedServicesPermitted/>
         </private>
      </privateData>
   </extensions>
</DeliveredEvent>''')
    # for i in a:
    #     m = parseBytes(i)
    #     print(m["monitorCrossRefID"])
    #     if m["monitorCrossRefID"] is None:
    #         print(m)
    for j in b:
        m = buildMessage(j, {"user": "12313213"}, eventid=2222)
        print(m["callID"])
        if m["deviceID"] and m["callID"]:
            # print(m["deviceID"], m["callID"])
            # m["deviceID"] = "user.deviceID"
            # m["callID"] = "user.callID"
            # print(m["deviceID"], m["callID"])
            pass
        print(m.get_all("callID"))
        # print("`````", m["callID"])
        # print("=====", m.body.get_all("deviceIdentifier"))
    last = buildMessage(j, {"user": "12313213"}, eventid=2222)["deviceObject"]
    te = """<?xml version="1.0" encoding="UTF-8"?>
<TransferedEvent xmlns="http://www.ecma-international.org/standards/ecma-323/csta/ed4">
<monitorCrossRefID>1184300014</monitorCrossRefID>
<primaryOldCall>
<callID>FF00010000000000CFB6706408000000</callID>
<deviceID>+17867104700</deviceID>
</primaryOldCall>
<transferringDevice>
<deviceIdentifier>+17867104700</deviceIdentifier>
</transferringDevice>
<transferredToDevice>
<deviceIdentifier>N&lt;+17867100001&gt;Line00001;displayNumber=00001</deviceIdentifier>
</transferredToDevice>
<transferredConnections>
<connectionListItem>
<newConnection>
<callID>0300010000000000CFB6706408000000</callID>
<deviceID>+17867100001</deviceID>
</newConnection>
<endpoint>
<deviceID>N&lt;+17867100001&gt;Line00001;displayNumber=00001</deviceID>
</endpoint>
</connectionListItem>
<connectionListItem>
<newConnection>
<callID>0300010000000000CFB6706408000000</callID>
<deviceID>N&lt;+17867100000&gt;;uid=FF00010000000000CFB6706408000000</deviceID>
</newConnection>
<endpoint>
<deviceID>N&lt;+17867100000&gt;+ 30(210)8216100;displayNumber=7867100000;uid=FF00010000000000CFB6706408000000</deviceID>
</endpoint>
</connectionListItem>
</transferredConnections>
<localConnectionInfo>null</localConnectionInfo>
<correlatorData>
<string>75726E3A6E656E613A7569643A63616C6C69643A62636631393235303838333532313330373135383A657362632E6E65743B75726E3A6E656E613A7569643A696E636964656E7469643A62636631393235303838333532313330373135383A657362632E6E6574</string>
</correlatorData>
<cause>singleStepTransfer</cause>
<servicesPermitted>
<callControlServices/>
<callAssociatedServices/>
<mediaAttachementServices/>
<routeingServices/>
<voiceServices/>
</servicesPermitted>
<mediaCallCharacteristics>
<mediaClass>
<voice>true</voice>
<image>false</image>
<message>false</message>
</mediaClass>
</mediaCallCharacteristics>
<callCharacteristics>
<priorityCall>true</priorityCall>
<highPriorityCall>true</highPriorityCall>
<sensitiveCall>true</sensitiveCall>
</callCharacteristics>
<extensions>
<privateData>
<private xmlns:scx="http://www.siemens.com/schema/csta">
<scx:extendedServicesPermitted/>
</private>
</privateData>
</extensions>
</TransferedEvent>"""
    tem = buildMessage(te)
    print(tem)
    print(tem.get_all("callID"))






"""\
Purpose: Visualize test scenarios
Initial Version: Costas Skarakis 12/20/2021
"""

import tkinter.tix as tix
import tkinter
import threading
import traceback
from time import ctime
from sip.SipEndpoint import SipEndpoint

MAXROWS = 40
MESSAGEENTRYWIDTH = 30


class LoadWindow:
    def __init__(self):
        self.root = tix.Tk()
        scroll_window = tix.ScrolledWindow(self.root, width=300, height=800, scrollbar=tix.AUTO)
        scroll_window.pack(expand=1, fill=tix.BOTH)
        self.window = scroll_window.window
        self.debug = tkinter.IntVar()
        self.debug_check = tkinter.Checkbutton(self.window, variable=self.debug, text="Enable Click error to debug")
        self.debug_check.grid(row=0, column=0, sticky="w")
        self.pack_legend()
        self.frames = {}
        self.subs = {}

    def pack_legend(self):
        legend = tix.Frame(self.window)
        legend.grid(row=1, column=0, stick="we")
        tix.Label(legend, text="Endpoint", width=12).pack(side=tix.LEFT)
        tix.Label(legend, text="State", width=5).pack(side=tix.LEFT)
        tix.Label(legend, text="SIP Message History", width=MESSAGEENTRYWIDTH - 1).pack(side=tix.LEFT)
        tix.Label(legend, text="Result", width=8).pack(side=tix.LEFT)

    def debug_window(self, number):
        if self.debug.get():
            frame = tkinter.Toplevel()
            frame.title(number)
            sub = self.subs[number]
            sub_frame = self.frames[number]
            message = ""
            if sub_frame.error_message.get():
                message += "========== EXCEPTION TEXT  ==========\n"
                message += sub_frame.error_message.get()
            message += "\n========== INCOMING SIP MESSAGE BUFFER  ==========\n"
            message += str(sub.message_buffer)
            message += "\n========== LAST_MESSAGES_PER_DIALOG ==========\n%s\n" % str(sub.last_messages_per_dialog)
            message += "\n========== DIALOGS ==========\n%s\n" % str(sub.dialogs)
            message += "\n========== REQUESTS ==========\n%s\n" % str(sub.requests)
            message += "\n========== ADDRESS: %s:%s==========\n" % (str(sub.ip), str(sub.port))
            text = tkinter.Text(frame)
            text.insert("end", message)
            text.pack(side="left", expand=1, fill="both")

    def paint(self, number):
        if number not in self.frames:
            count = len(self.frames) + 2
            sub_row = count % MAXROWS
            sub_col = count // MAXROWS
            frame = tix.Frame(self.window)
            frame.grid(row=sub_row, column=sub_col)
            frame.label = tix.Label(frame, text=number, width=13)
            frame.arrow = tix.Label(frame, width=3, bg="yellow")
            frame.messagevar = tkinter.StringVar()
            frame.message = tkinter.Entry(frame, width=MESSAGEENTRYWIDTH, fg="black", exportselection=True,
                                          selectforeground='blue', state="readonly",
                                          textvariable=frame.messagevar)
            frame.error = tix.Label(frame, width=8)
            frame.error_message = tkinter.StringVar()
            frame.label.pack(side=tix.LEFT)
            frame.arrow.pack(side=tix.LEFT)
            frame.message.pack(side=tix.LEFT)
            frame.error.pack(side=tix.LEFT)
            frame.error.bind('<Button-1>', lambda event, n=number: self.debug_window(n))
            self.frames[number] = frame

    def arrow(self, number, text):
        self.frames[number].arrow["text"] = text

    def message(self, number, text):
        entry_text = self.frames[number].messagevar.get()
        entry_text += text
        self.frames[number].messagevar.set(entry_text)
        self.frames[number].message.xview_moveto(1)

    def error(self, number, text, exception_text):
        self.frames[number].error["text"] = text
        self.frames[number].error_message.set("%s:\n%s\n" % (ctime(),exception_text))
        self.frames[number].error["bg"] = "tomato"

    def no_error(self, number):
        self.frames[number].error["text"] = chr(10004)
        self.frames[number].error["bg"] = "green3"

    def start(self, func_on_thread, *args, **kwargs):
        func_thread = threading.Thread(target=func_on_thread, args=args, kwargs=kwargs).start()
        self.root.mainloop()
        print("Exit main loop")
        func_thread.join()


class SipEndpointView(SipEndpoint):
    def __init__(self, view, number):
        self.view = view
        self.view.subs[number] = self
        super().__init__(number)

    def connect(self, *args, **kwargs):
        self.view.paint(self.number)
        super().connect(*args, **kwargs)

    def use_link(self, *args, **kwargs):
        self.view.paint(self.number)
        super().use_link(*args, **kwargs)

    def send(self, *args, **kwargs):
        self.view.arrow(self.number, "Sndg")
        message = super().send(*args, **kwargs)
        self.view.arrow(self.number, "Sent")
        # if "expected_response" in kwargs and kwargs["expected_response"]:
        #     label = "<-" + kwargs["expected_response"][:3]
        #     self.view.message(self.number, label)
        return message

    def reply(self, *args, **kwargs):
        self.view.arrow(self.number, "Sndg")
        message = super().reply(*args, **kwargs)
        self.view.arrow(self.number, "Sent")
        label = "->" + message.get_status_or_method()[:3]
        self.view.message(self.number, label)
        return message

    def send_new(self, *args, **kwargs):
        self.view.arrow(self.number, "Sndg")
        message = super().send_new(*args, **kwargs)
        self.view.arrow(self.number, "Sent")
        label = "+->%s(%s)" % (message.get_status_or_method()[:3], message["Call-ID"])
        self.view.message(self.number, label)
        # if "expected_response" in kwargs and kwargs["expected_response"]:
        #     label = "<-" + kwargs["expected_response"][:3]
        #     self.view.message(self.number, label)
        return message

    def wait_for_message(self, message_type, **kwargs):
        self.view.arrow(self.number, "Wng")
        self.view.message(self.number, "<-")
        error = True
        try:
            message = super().wait_for_message(message_type, **kwargs)
            error = False
            self.view.arrow(self.number, "Rcvd")
            label = message.get_status_or_method()[:3]
            self.view.message(self.number, label)
            return message
        finally:
            if error:
                self.view.error(self.number, "No %s" % message_type[:3], traceback.format_exc())
            else:
                self.view.no_error(self.number)

    def colour(self, colour):
        self.view.frames[self.number].arrow["bg"] = colour

    def update_text(self, text=""):
        self.view.message(self.number, text)

    def update_arrow(self, arrow=""):
        self.view.arrow(self.number, arrow)

    def make_busy(self, busy=True):
        super().make_busy(busy)
        self.colour(["yellow", "cyan"][busy])
        self.view.arrow(self.number, ["Idle", "Busy"][busy])
